import copy

import matplotlib

matplotlib.use("WXAgg")

from matplotlib import __version__ as matplotlibversion

import wx

if wx.__version__ < "4.":
    WXPYTHON4 = False
else:
    WXPYTHON4 = True
    wx.OPEN = wx.FD_OPEN

# --- ---------------  Plot limits board  parameters
class PlotLimitsBoard(wx.Dialog):
    """
    Class to set limits parameters of plot

    parent class must have
    xlim, ylim,  flipyaxis attributes
    _replot(), getlimitsfromplot() methods
    """
    def __init__(self, parent, _id, title, data_dict):
        """
        initialize board window
        """
        wx.Dialog.__init__(self, parent, _id, title, size=(400, 250))

        self.parent = parent

        self.data_dict = data_dict

        xlim = self.data_dict["xlim"]
        ylim = self.data_dict["ylim"]

        self.init_xlim = copy.copy(xlim)
        self.init_ylim = copy.copy(ylim)

        self.xmin, self.xmax, self.ymin, self.ymax = None, None, None, None
        self.xlim, self.ylim = None, None

        font3 = wx.Font(10, wx.MODERN, wx.NORMAL, wx.BOLD)
        txt1 = wx.StaticText(self, -1, "X and Y limits controls")
        txt2 = wx.StaticText(self, -1, "Data type: %s" % self.data_dict["datatype"])

        txt1.SetFont(font3)

        self.txtctrl_xmin = wx.TextCtrl(self, -1, str(xlim[0]), style=wx.TE_PROCESS_ENTER)
        self.txtctrl_xmax = wx.TextCtrl(self, -1, str(xlim[1]), style=wx.TE_PROCESS_ENTER)
        self.txtctrl_ymin = wx.TextCtrl(self, -1, str(ylim[0]), style=wx.TE_PROCESS_ENTER)
        self.txtctrl_ymax = wx.TextCtrl(self, -1, str(ylim[1]), style=wx.TE_PROCESS_ENTER)

        self.txtctrl_xmin.Bind(wx.EVT_TEXT_ENTER, self.onEnterValue)
        self.txtctrl_xmax.Bind(wx.EVT_TEXT_ENTER, self.onEnterValue)
        self.txtctrl_ymin.Bind(wx.EVT_TEXT_ENTER, self.onEnterValue)
        self.txtctrl_ymax.Bind(wx.EVT_TEXT_ENTER, self.onEnterValue)

        fittodatabtn = wx.Button(self, -1, "Fit to Data")
        fittodatabtn.Bind(wx.EVT_BUTTON, self.onFittoData)

        acceptbtn = wx.Button(self, -1, "Accept")
        cancelbtn = wx.Button(self, -1, "Cancel")

        acceptbtn.Bind(wx.EVT_BUTTON, self.onAccept)
        cancelbtn.Bind(wx.EVT_BUTTON, self.onCancel)

        if WXPYTHON4:
            grid = wx.GridSizer(5, 10, 10)
        else:
            grid = wx.GridSizer(6, 5)

        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, "Y"), wx.ALIGN_CENTER_HORIZONTAL)
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))

        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, "MAX"), wx.ALIGN_CENTER_HORIZONTAL)
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))

        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(self.txtctrl_ymax)
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))

        grid.Add(wx.StaticText(self, -1, "X    min"), wx.ALIGN_RIGHT)
        grid.Add(self.txtctrl_xmin)
        grid.Add(fittodatabtn)
        grid.Add(self.txtctrl_xmax)
        grid.Add(wx.StaticText(self, -1, "MAX"))

        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(self.txtctrl_ymin)
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))

        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, "min"), wx.ALIGN_CENTER_HORIZONTAL)
        grid.Add(wx.StaticText(self, -1, ""))
        grid.Add(wx.StaticText(self, -1, ""))

        btnssizer = wx.BoxSizer(wx.HORIZONTAL)
        btnssizer.Add(acceptbtn, 0, wx.ALL)
        btnssizer.Add(cancelbtn, 0, wx.ALL)

        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(txt1)
        vbox.Add(txt2)
        vbox.Add(grid)
        vbox.Add(btnssizer)

        self.SetSizer(vbox)

    def onEnterValue(self, _):
        self.readvalues()
        self.updateplot()

    def onFittoData(self, _):

        xmin = self.data_dict["dataXmin"]
        xmax = self.data_dict["dataXmax"]
        ymin = self.data_dict["dataYmin"]
        ymax = self.data_dict["dataYmax"]

        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax

        self.txtctrl_xmin.SetValue(str(self.xmin))
        self.txtctrl_xmax.SetValue(str(self.xmax))
        self.txtctrl_ymin.SetValue(str(self.ymin))
        self.txtctrl_ymax.SetValue(str(self.ymax))

        self.setxylim()

        self.updateplot()

    def updateplot(self):
        self.parent.xlim = self.xlim
        self.parent.ylim = self.ylim
        self.parent.getlimitsfromplot = False
        self.parent._replot()

    def readvalues(self):
        self.xmin = float(self.txtctrl_xmin.GetValue())
        self.xmax = float(self.txtctrl_xmax.GetValue())
        self.ymin = float(self.txtctrl_ymin.GetValue())
        self.ymax = float(self.txtctrl_ymax.GetValue())

        self.setxylim()

    def setxylim(self):
        self.xlim = (self.xmin, self.xmax)

        if self.parent.flipyaxis is not None:
            if not self.parent.flipyaxis:
                self.ylim = (self.ymin, self.ymax)
            else:
                # flip up for marccd roper...
                self.ylim = (self.ymax, self.ymin)
        else:
            self.ylim = (self.ymin, self.ymax)

    def onAccept(self, _):

        self.readvalues()
        self.updateplot()

        self.Close()

    def onCancel(self, _):

        self.parent.xlim = self.init_xlim
        self.parent.ylim = self.init_ylim
        self.Close()



import sys, os

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import findorient as FO
import generaltools as GT
import CrystalParameters as CP
import LaueGeometry as LTGeo
from dict_LaueTools import dict_Materials


def Test_Build_and_Query_LUT(
    angle=69, tol=1, n=3, latticeparams=(2, 1, 4, 75, 90, 120)):
    """
    Test_Build_and_Query_LUT
    """
    # compute LUT outside loop:
    hkl_all = GT.threeindices_up_to(n, remove_negative_l=CP.isCubic(latticeparams))

    if 1:  # filterharmonics:
        #        hkl_all = CP.FilterHarmonics_2(hkl_all)
        hkl_all = FO.FilterHarmonics(hkl_all)

    Gstar_metric = CP.Gstar_from_directlatticeparams(*latticeparams)
    # GenerateLookUpTable
    LUT = FO.GenerateLookUpTable(hkl_all, Gstar_metric)

    sol69 = FO.QueryLUT(LUT, angle, tol)

    print("solutions", sol69)

    return sol69


def test_FilterHarmonics():
    """
    test harmonics removal
    """
    a = np.arange(3 * 8).reshape((8, 3))
    a[-1] = np.array([-6, -8, -10])
    a[-3] = np.array([9, 12, 15])
    fila = FO.FilterHarmonics(a)
    print("a", a)
    print("filtered a")
    print(fila)

    print("\n other test")
    testhkl = np.array(
        [
            [-2, -4, 2],
            [0, -2, 2],
            [-3, -11, 3],
            [-1, -11, 3],
            [-5, -9, 3],
            [-3, -9, 3],
            [-1, -9, 3],
            [1, -9, 3],
            [-5, -7, 3],
            [-3, -7, 3],
            [-1, -7, 3],
            [1, -7, 3],
            [-5, -5, 3],
            [-3, -5, 3],
            [-1, -5, 3],
            [1, -5, 3],
            [-3, -3, 3],
            [-1, -3, 3],
            [1, -3, 3],
            [-2, -10, 4],
            [-4, -8, 4],
            [0, -8, 4],
            [-6, -6, 4],
            [-2, -6, 4],
            [2, -6, 4],
            [0, -4, 4],
            [-2, -2, 4],
            [-5, -7, 5],
        ]
    )
    print("testhkl")
    print(testhkl)
    print("one of  [  0,  -2,   2] and [  0,  -4,   4] must be removed")

    print(FO.FilterHarmonics(testhkl))


def test():
    """
    test
    """
    # GT.find_closest(input_array, target_array, tol)

    # datafrompickle=open('recognised.rec','r')
    # reconnu = pickle.load(datafrompickle)
    # datafrompickle.close()

    recognised = [
        [72.263978989400002, 19.5486315279, 111, 12],
        [109.63069847600001, -24.477560408599999, 110, 16],
        [95.941622918999997, 10.5435782548, 210, 21],
        [82.3329445217, 4.2357560583199998, 311, 31],
        [85.062674650700004, -3.4110763979200001, 210, 32],
        [97.118855973999999, -0.87674472922199997, 211, 64],
        [106.209401235, -15.1071780961, 311, 93],
        [49.406572447000002, -12.219836147100001, 311, 133],
        [93.0858676231, 3.5826456916599998, 221, 147],
    ]

    recog = recognised  # reconnu

    print("-----------------------------------------")
    print("First two recognised spots Matrix proposition")
    indexpoint1 = 0  # default = 0
    indexpoint2 = 1  # default = 1
    ang = 0.1
    print(
        "For spots #%d and #%d in data"
        % (recog[indexpoint1][-1], recog[indexpoint2][-1])
    )
    print("Angular tolerance: %.2f deg" % ang)
    FO.proposematrix(indexpoint1, indexpoint2, recog, ang)
    print("------------------------\n")

    print("Now looking at all spots pairs distances and compare with lookup table")
    toleranceang = 0.05
    print("-----------------------------------------")
    print("Angular tolerance for multiple search: %.2f" % toleranceang)
    result = FO.Allproposedmatrix(recog, toleranceang)
    # print result

    if len(result) > 0:
        print("Cool, I can propose %s matrix" % len(result))
        for elem in result:
            spotindices = elem[0]
            print(
                "With spots #%d and #%d in data"
                % (recog[spotindices[0]][-1], recog[spotindices[1]][-1])
            )
            print(elem[1])

        print("Now removing similar matrix if needed")
        singlematrixlist = [result[0][1]]
    else:
        print("Argh no matrix found!")

    for j in list(range(1, len(result))):
        listcorrel = [
            np.corrcoef(np.ravel(result[j][1]), np.ravel(singlematrixlist[k]))[0][1]
            for k in range(len(singlematrixlist))
        ]

        booltable = np.array(listcorrel) < 0.995  # True if matrices are uncorrelated

        if all(booltable):  # if matrix proposed is different from the others

            singlematrixlist.append(result[j][1])

    print("Similar matrices removed")
    print(singlematrixlist)

    print("TODO: looking at same matrix by frame permutation")


def example_Neighbourslist():
    """ 
    give list of planes that are at 25 deg +- tol from a the 210 plane
    """
    print("Planes that are distant from [2,1,0] plane from 20 to 30 degree")
    tol = 5.0
    table_210 = FO.prodtab_210()
    pos_close = GT.find_closest(array([25.0]), table_210, tol)
    plane_sol = np.take(FO.LISTNEIGHBORS_210, pos_close[1], axis=0)
    distance_sol = np.take(table_210, pos_close[1])
    for plane, dist in zip(plane_sol, distance_sol):
        print("plane %s at %.2f degree from plane [2,1,0]" % (plane, dist))


def test_PlanePairs_from2sets():
    latticeparams = dict_Materials["Cu"][1]
    n = 3
    # compute LUT outside loop:
    hkl_all = GT.threeindices_up_to(n, remove_negative_l=CP.isCubic(latticeparams))

    if 1:  # filterharmonics:
        #        hkl_all = CP.FilterHarmonics_2(hkl_all)
        hkl_all = FO.FilterHarmonics(hkl_all)

    Gstar = CP.Gstar_from_directlatticeparams(*latticeparams)

    hkl1 = [[1, 0, 0], [2, 2, 1]]
    hkl2 = hkl_all

    FO.PlanePairs_from2sets(
        25.35, 5, hkl1, hkl2, Gstar, onlyclosest=1, filterharmonics=1, verbose=1
    )

    FO.PlanePairs_from2sets(
        25.35, 5, hkl1, hkl2, Gstar, onlyclosest=0, filterharmonics=1, verbose=1
    )


#     Test_Build_and_Query_LUT(angle=69, tol=1, n=4, latticeparams=(2, 1, 4, 75, 90, 120))


# test() # old test with data from recognition.py


#     PlanePairs_from2sets(25.1, .5,
#                             hkl1, hkl2, Gstar,
#                             onlyclosest=1, filterharmonics=1, verbose=1)

if 1:  # some tests to play with angles LUT

    # these parameters cannot be randomly chosen (they are related each other)
    # But by setting one angle to 90. you may play a lot with the other lattice parameters
    a, b, c, alpha, beta, gamma = 5.0, 2.3, 4.86, 89.56, 68.25, 90.0
    # a,b,c,alpha,beta,gamma = 4.99,4.99,17.061,90,90,120 # trigonal, 'hexagonal' lattice

    # compute metric tensor
    Gstar = CP.Gstar_from_directlatticeparams(a, b, c, alpha, beta, gamma)

    print("Gstar", Gstar)

    # HKL_list = np.array([[0,0,1],[1,0,4],[-1,1,4],[1,2,0]])
    # for hkl1 in HKL_list[:-1]:
    # for hkl2 in HKL_list[1:]:
    # angle = CP.AngleBetweenTwoNormals(hkl1, hkl2 ,Gstar)
    # print "<%s , %s> = %.5f"%(str(hkl1),str(hkl2),angle)

    # some hkl lists
    hkl1 = GT.Positiveindices_up_to(3)
    hkl2 = GT.Positiveindices_up_to(2)

    print("hkl2", hkl2)
    # build LUT
    LUT = FO.GenerateLookUpTable_from2sets(hkl1, hkl2, Gstar)

    print("LUT", LUT)

    # test_hkl = np.array([[5,-5,3],[203,77,91],[-4,2,1],[6,3,3],[8,4,2],[1,1,1],[2,2,2],[0,2,8]])
    # GCDs = GT.GCD(np.abs(test_hkl))

    # res = GT.reduceHKL(test_hkl)

    pp, LUT = FO.PlanePairs_from2sets(
        12.58, 1.0, hkl1, hkl2, "Cu", onlyclosest=1, filterharmonics=1, verbose=1
    )

if 0:  # test of orient matrix from distorted unit cell

    from numpy import *

    a, b, c, alpha, beta, gamma = 3.2, 2.5, 5.0, 86, 90.0, 120.0
    latticeparams = a, b, c, alpha, beta, gamma
    B = CP.calc_B_RR([a, b, c, alpha, beta, gamma])
    # metric tensor ---------
    Gstar_metric = CP.Gstar_from_directlatticeparams(a, b, c, alpha, beta, gamma)

    # hkl = np.array([[-2,0,2],[-6,0,2], [-3,1,1],[-2,4,2]])
    # matorient = eye(3)

    hkl = np.array([[3, 1, 7], [1, 1, 3], [2, 2, 2], [5, 1, 1]])
    matorient = array(
        [
            [-0.6848285, 0.2544487, -0.6828365],
            [-0.7064087, -0.0017916, 0.7078019],
            [0.1788759, 0.9670846, 0.1809717],
        ]
    )  # surf 111

    # Digest of Laue6.py --------------------
    # G* taken into account distorsion
    Gstarprime = dot(B, hkl.T).T

    print("matrorient for simulation", matorient)

    q = dot(matorient, Gstarprime.T).T

    qn = sqrt(sum(q ** 2, axis=1))

    uq = q / reshape(qn, (len(qn), 1))

    # ui in Lauetools frame
    ui = array([1, 0, 0.0])
    # uf = ui -2(ui.uq)uq

    coef = -2.0 * dot(ui, uq.T)
    uf = ui + uq * reshape(coef, (len(qn), 1))

    # kf = ( - sin 2theta sin chi, cos 2theta  , sin 2theta cos chi) XMAS
    # kf = ( cos 2theta, sin 2theta sin chi,   , sin 2theta cos chi) LT

    EPS = 0.00000001
    PI = np.pi
    # in LT
    chi = (180.0 / PI) * arctan(uf[:, 1] / (uf[:, 2] + EPS))
    twtheta = (180.0 / PI) * arccos(uf[:, 0])

    coord = array([twtheta, chi]).T
    # ---------------------------------------------------------------

    # coord = [[90,0],[143.1301,0],[139.1214,    26.5651],[70.5288,    45.0000]]

    hkl1 = hkl[0]
    coord1 = coord[0] + array([0.1, -1])  # add some noise

    hkl2 = hkl[2]
    coord2 = coord[2] + array([-0.05, 0.7])  # add some noise

    mat = FO.OrientMatrix_from_2hkl(
        hkl1, coord1, hkl2, coord2, B, verbose="yes", frame="lauetools"
    )

    # --------------------------------------------------
    # check from all kinds of situation

    # res_mat = []

    # for i in range(len(hkl)-1):
    # for j in range(i+1,len(hkl)):

    # hkl1 = hkl[i]
    # hkl2 = hkl[j]

    # coord1 = coord[i] +array([0.1,-1]) # add some noise
    # coord2 = coord[j] +array([-.05,.7])

    # mat = OrientMatrix_from_2hkl(hkl1, coord1, hkl2, coord2, B, verbose='yes', frame = 'lauetools')

    # res_mat.append(mat)
    # print "\n\nmatorient",mat
    # print "*****************\n"

    # --------------------------------------------------

    if 0:
        # build linear matrix equation to solve with pinv
        # for m = 1,2,3
        # we look for operator that solve
        # qm = (U) Gm*
        # size(qm) = size(Gm*) = 3
        # shape (U) = (3,3)
        # let U = A whose element are a ij
        # the three equations (qix,qiy,qiz) = ((a11,a12,a13),(a21,a22,a23),(a31,a32,a33))(Gi*x,Gi*y,Gi*z)
        # can be rearranged like M P = Q
        # where P is 9 element of A: [a11,a12,a13,a21,a22,a23,a31,a32,a33]
        #         Q is a stack of the sthree components of each qm
        #       Q = [q1x,q1y,q1z,q2x,q2y,q2z,q3x,q3y,q3z]
        # and M is a quite sparse matrix containing row of Gm* elements
        #     M is a 3 times vertically stacked pattern:
        #  patternM =[[G1x*,G1y*,G1z*,0,0,0,0,0,0],[0,0,0,G2x*,G2y*,G2z*,0,0,0],[0,0,0,0,0,0,G3x*,G3y*,G3z*]]

        # in case of distorted unit cell: Gm* are unit vector from B Gm*
        # qm are also unit vectors

        G1 = np.dot(B, np.array(hkl1))
        G2 = np.dot(B, np.array(hkl2))
        G3 = np.cross(G1, G2)

        normG1 = np.sqrt(np.dot(G1, G1))
        normG2 = np.sqrt(np.dot(G2, G2))
        normG3 = np.sqrt(np.dot(G3, G3))

        print("norms of G1,G2,G3", normG1, normG2, normG3)

        matrice_P = np.array([G1 / normG1, G2 / normG2, G3 / normG3])

        # building M
        patternM = np.zeros((3, 9))
        patternM[0, :3] = matrice_P[0]
        patternM[1, 3:6] = matrice_P[1]
        patternM[2, 6:] = matrice_P[2]

        M = np.vstack((patternM, patternM, patternM))

        # similar matrix but with 2theta,chi coordinate
        twicetheta_1 = coord1[0]
        chi_1 = coord1[1]
        twicetheta_2 = coord2[0]
        chi_2 = coord2[1]

        # expression of unit q in chosen frame
        frame = "lauetools"
        qq1 = LTGeo.unit_q(twicetheta_1, chi_1, frame=frame)
        qq2 = LTGeo.unit_q(twicetheta_2, chi_2, frame=frame)

        qq3prod = np.cross(
            qq1, qq2
        )  # can be negative, we want to have finally a matrix with one eigenvalue of +1
        qq3n = np.sqrt(np.dot(qq3prod, qq3prod)) * 1.0

        Q = np.ravel(np.array([qq1, qq2, qq3prod / qq3n]))

        # solving for P : MP= Q -----------------------
        # 1-  use SVD:
        U, s, V = np.linalg.svd(M)
        # then P   =  V diag(1/s) U.T Q
        # with diag(1/s) is a diagonal matrix made of inverse of singular values
        # if one singular values is very close to 0 then the inverse of this value must replace by zero!!
        print("singular values", s)

        threshold = 0.001
        invs = np.where(s < threshold, 0, 1.0 / s)

        print("inverse singular value", invs)
        invS = np.diag(invs)

        invM = np.dot(V, np.dot(invS, U.T))

        P = np.dot(invM, Q)



r"""
.. module:: matchingrate Documentation
    :synopsis: module to compute matching figure from two sets of laue spots properties

.. moduleauthor:: JS Micha, 2019,  micha 'at' esrf 'dot'fr
"""
import copy
import sys
from numpy import array, where, argmin, amin, mean
import numpy as np

SCIKITLEARN = True
try:
    import sklearn.metrics as sm
except ImportError:
    print("sklearn.metrics is not installed !! ")
    SCIKITLEARN = False


if sys.version_info.major == 3:

    from . import lauecore as LAUE
    from . import CrystalParameters as CP
    from . import generaltools as GT
    from . import LaueGeometry as LaueGeo
    from . dict_LaueTools import CST_ENERGYKEV
    from . dict_LaueTools import dict_Materials
    from . import indexingSpotsSet as ISS

else:
    import lauecore as LAUE
    import CrystalParameters as CP
    import generaltools as GT
    import LaueGeometry as LaueGeo
    from dict_LaueTools import CST_ENERGYKEV
    from dict_LaueTools import dict_Materials
    import indexingSpotsSet as ISS


try:
    if sys.version_info.major == 3:
        from . import angulardist
    else:
        import angulardist

    USE_CYTHON = True
except ImportError:
    # print("-- warning. Cython compiled 'angulardist' module for fast computation of "
    #                                                         "angular distance is not installed!")
    # print("Using default module")
    USE_CYTHON = False


# --- ------------ CONSTANTS
DEG = np.pi / 180.0

#------ USE_CYTHON= FALSE ----------------
def getArgmin(tab_angulardist):
    r"""
    temporarly doc
    from matrix of mutual angular distances return index of closest neighbour

    .. todo:: to explicit documentation as a function of tab_angulardist properties only
    """
    return argmin(tab_angulardist, axis=1)


def SpotLinks(twicetheta_exp,
                chi_exp,
                dataintensity_exp,  # experimental data
                veryclose_angletol,  # tolerance angle
                twicetheta,  # theoretical angles
                chi,
                Miller_ind,
                energy,  # theoretical data
                absoluteindex=None,
                verbose=0):
    r"""
    Creates automatically links between close experimental and theoretical spots
    in 2theta, chi angles (kf) coordinates

    :param twicetheta_exp: list of exp. 2theta angles
    :param chi_exp: list of exp. chi angles
    :param dataintensity_exp: list of dataintensity_exp

    :param veryclose_angletol: finest tolerance angle for association in degree

    :param twicetheta: list of theoretical 2theta angles
    :param chi: list of theoretical chi angles
    :param Miller_ind: list of theoretical 3D Miller_ind
    :param energy: list of theoretical spot energies (keV)

    :param absoluteindex: list of absolute exp. spot index to output in res
                            (conversion table between relative and absolute exp. spot index)
                           i.e.  absoluteindex[localindex] = absolute index

                        if absoluteindex = None: spots indices and order are those of data
                                    (twicetheta_exp, chi_exp, dataintensity_exp)
                    = list_of_absolute_indices
                            : list containing the absolute indices

    :returns: * refine_indexed_spots: dict. with key= exp. spotindex and val=[exp. spotindex,h,k,l]
            * linkedspots_link: list of [absolute exp. spotindex, theo_id]
            * linkExpMiller_link: list of [absolute exp. spotindex, h,k,l]
            * linkIntensity_link: list of [exp. spot intensity]
            * linkResidues_link: list of [absolute exp. spotindex, theo. spotindex, angular pair distance (degrees)]
            * linkEnergy_link: list of [absolute exp. spotindex, theo. spotindex, theo. spot energy (keV)]
            * fields: ["#Spot Exp", "#Spot Theo", "h", "k", "l", "Intensity", "residues (deg)"]

            linkResidues.append([absoluteexpspotindex, theo_id, Resi[theo_id]])
            linkEnergy.append([absoluteexpspotindex, theo_id, Energy_id])
            linkIntensity.append(dataintensity_exp[exp_id])


    .. todo:: To improve! , always difficult to understand immediately the algorithm :-[
    """

    if verbose:
        print("\n ************** Spots Association ************** \n")

    Resi, ProxTable = getProximity(np.array([twicetheta, chi]),  # warning array(2theta, chi)
                            twicetheta_exp / 2.0,
                            chi_exp,  # warning theta, chi for exp
                            proxtable=1,
                            angtol=veryclose_angletol,
                            verbose=0,
                            signchi=1,
                        )[:2]  # sign of chi is +1 when apparently SIGN_OF_GAMMA=1

    # ProxTable is table giving the closest exp.spot index for each theo. spot
    # len(Resi) = nb of theo spots
    # len(ProxTable) = nb of theo spots
    # ProxTable[index_theo]  = index_exp   closest link
    # print "Resi",Resi
    # print "ProxTable",ProxTable
    # print "Nb of theo spots", len(ProxTable)

    # array of theo. spot index
    very_close_ind = np.where(Resi < veryclose_angletol)[0]
    # print "In OnLinkSpotsAutomatic() very close indices",very_close_ind

    # build a list all theo. spots such as: exp_index = List_Exp_spot_close[theo_index]
    List_Exp_spot_close = []
    Miller_Exp_spot = []
    Energy_Exp_spot = []

    # if nb of pairs between exp and theo spots is enough
    if len(very_close_ind) > 0:
        # loop over theo spots index
        for theospot_ind in very_close_ind:
            # print "theospot_ind ",theospot_ind
            List_Exp_spot_close.append(ProxTable[theospot_ind])
            Miller_Exp_spot.append(Miller_ind[theospot_ind])
            Energy_Exp_spot.append(energy[theospot_ind])

    #    print "List_Exp_spot_close", List_Exp_spot_close
    #    print "Miller_Exp_spot", Miller_Exp_spot

    if not List_Exp_spot_close:
        print("Found no pair within tolerance in SpotLinks()")
        return 0

    elif len(List_Exp_spot_close) == 1:
        print("Just a single found! Nb of pairs equal to 1  is not implemented yet in SpotLinks()")
        return 0

    # --------------------------------------------------------------
    # removing exp spot which appears many times
    # (close to several simulated spots of one grain)--------------
    #   now len(List_Exp_spot_close) >= 2
    # ---------------------------------------------------------

    arrayLESC = np.array(List_Exp_spot_close, dtype=float)

    sorted_LESC = np.sort(arrayLESC)

    # print "List_Exp_spot_close", List_Exp_spot_close
    # print "sorted_LESC", sorted_LESC

    diff_index = sorted_LESC - np.array(list(sorted_LESC[1:]) + [sorted_LESC[0]])
    toremoveindex = np.where(diff_index == 0)[0]

    # print "toremoveindex", toremoveindex

    # print "number labelled exp spots", len(List_Exp_spot_close)
    # print "List_Exp_spot_close", List_Exp_spot_close
    # print "Miller_Exp_spot", Miller_Exp_spot

    # find exp spot that can not be indexed safely
    # (too many theo. neighbouring spots)
    if len(toremoveindex) > 0:
        # index of exp spot in arrayLESC that are duplicated
        TOLERANCEANGLE = 0.1
        ambiguous_exp_ind = GT.find_closest(
            np.array(sorted_LESC[toremoveindex], dtype=float), arrayLESC, TOLERANCEANGLE
        )[1]
        # print "ambiguous_exp_ind", ambiguous_exp_ind

        # tagging (inhibiting) exp spots
        # that belong ambiguously to several simulated grains
        for exp_ind in ambiguous_exp_ind:
            Miller_Exp_spot[exp_ind] = None
            Energy_Exp_spot[exp_ind] = 0.0

    # -----------------------------------------------------------
    ProxTablecopy = copy.copy(ProxTable)
    # tag duplicates in ProxTable with negative sign ----------------------
    # ProxTable[index_theo] = index_exp   closest link

    for theo_ind, exp_ind in enumerate(ProxTable):
        where_th_ind = np.where(ProxTablecopy == exp_ind)[0]
        # print "theo_ind, exp_ind ******** ",theo_ind, exp_ind
        if len(where_th_ind) > 1:
            # exp spot (exp_ind) is close to several theo spots
            # then tag the index with negative sign
            for indy in where_th_ind:
                ProxTablecopy[indy] = -ProxTable[indy]
            # except that which corresponds to the closest
            closest = np.argmin(Resi[where_th_ind])
            # print "residues = Resi[where_th_ind]",Resi[where_th_ind]
            # print "closest",closest
            # print "where_exp_ind[closest]",where_th_ind[closest]
            # print "Resi[where_th_ind[closest]]", Resi[where_th_ind[closest]]
            ProxTablecopy[where_th_ind[closest]] *= -1

    # ------------------------------------------------------------------
    # print "ProxTable after duplicate removal tagging"
    # print ProxTablecopy

    # print "List_Exp_spot_close",List_Exp_spot_close
    # print "Results",[Miller_Exp_spot, List_Exp_spot_close]

    # list of exp. spot index that have only one theo. neighbouring spots
    singleindices = []

    # dictionary of links (pairs) between exp. and theo. spots
    refine_indexed_spots = {}

    # loop over theo. spot index
    for theo_ind in list(range(len(List_Exp_spot_close))):

        exp_index = List_Exp_spot_close[theo_ind]

        # print "exp_index",exp_index

        # there is not exp_index in singleindices
        if not singleindices.count(exp_index):
            # so append singleindices with exp_index
            singleindices.append(exp_index)

            # lisf of index of theo spot that are the closest for a given exp. spot
            theo_index = np.where(ProxTablecopy == exp_index)[0]
            # print "theo_index", theo_index

            # unambiguous pairing
            if len(theo_index) == 1:
                refine_indexed_spots[exp_index] = [exp_index,
                                                    theo_index,
                                                    Miller_Exp_spot[theo_ind],
                                                    Energy_Exp_spot[theo_ind]]

            # in case of several theo. candidate, keep the closest to exp. spot
            else:
                # print "Resi[theo_index]", Resi[theo_index]
                closest_theo_ind = np.argmin(Resi[theo_index])
                # print theo_index[closest_theo_ind]

                # pairing if distance withing angular tolerance
                if Resi[theo_index][closest_theo_ind] < veryclose_angletol:
                    refine_indexed_spots[exp_index] = [
                        exp_index,
                        theo_index[closest_theo_ind],
                        Miller_Exp_spot[theo_ind],
                        Energy_Exp_spot[theo_ind]]

        # there is already exp_index in singleindices
        else:
            # do not update the dictionary 'refine_indexed_spots'
            if verbose:
                print("Experimental spot #%d may belong to several theo. spots!"
                    % exp_index)
    # find theo spot linked to exp spot ---------------------------------

    # refine_indexed_spots is a dictionary:
    # key is experimental spot index and value is [experimental spot index,h,k,l]
    # print "refine_indexed_spots",refine_indexed_spots

    listofpairs = []
    linkExpMiller = []
    linkIntensity = []
    linkResidues = []
    linkEnergy = []
    # Dataxy = []

    for val in list(refine_indexed_spots.values()):

        exp_id, theo_id, Miller_id, Energy_id = val

        if Miller_id is not None:

            # absoluteindex is a list of absolute exp. spots indices
            if absoluteindex is not None:
                absoluteexpspotindex = absoluteindex[exp_id]
            else:
                absoluteexpspotindex = exp_id

            # appending lists of links
            # Exp, Theo,  where -1 for specifying that it came from automatic linking
            listofpairs.append([absoluteexpspotindex, theo_id])
            linkExpMiller.append(
                [float(absoluteexpspotindex)] + [float(elem) for elem in Miller_id])  # float(val) for further handling as floats array
            linkResidues.append([absoluteexpspotindex, theo_id, Resi[theo_id]])
            linkEnergy.append([absoluteexpspotindex, theo_id, Energy_id])
            linkIntensity.append(dataintensity_exp[exp_id])
            # Dataxy.append([ LaueToolsframe.data_pixX[val[0]], LaueToolsframe.data_pixY[val[0]]])

    linkedspots_link = np.array(listofpairs)
    linkExpMiller_link = linkExpMiller
    linkIntensity_link = linkIntensity
    linkResidues_link = linkResidues
    linkEnergy_link = linkEnergy
    fields = ["#Spot Exp", "#Spot Theo", "h", "k", "l", "Intensity", "residues (deg)"]

    # self.Data_X, self.Data_Y = np.transpose( np.array(Dataxy) )

    return (refine_indexed_spots,
        linkedspots_link,
        linkExpMiller_link,
        linkIntensity_link,
        linkResidues_link,
        linkEnergy_link,
        fields)

def getProximity_multimatrices(Arr_Theo2Theta, Arr_TheoChi, data_theta, data_chi, angtol=0.5,
                                                                            proxtable=0,
                                                                            verbose=0,
                                                                            signchi=1,
                                                                            usecython=USE_CYTHON):
    """
    WARNING: ArrTheo2theta   contains 2theta instead of data_theta contains theta !
    """
    nbpeaksbunches = len(Arr_Theo2Theta)
    # theo simul data
    # theodata = array([np.ravel(Arr_Theo2Theta)/2.0, signchi *np.ravel(Arr_TheoChi)]).T
    # # exp data
    sorted_data = array([data_theta, data_chi]).T

    #     table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)
    #     print "table_dist_old", table_dist[:5, :5]
    #     print "table_dist_old", table_dist[-5:, -5:]
    #     print "table_dist_old", table_dist.shape

    print('sorted_data.shape', sorted_data.shape)
    print('Arr_Theo2Theta.shape', Arr_Theo2Theta.shape)

    if not usecython:
        # table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)

        table_dist = GT.computeMutualAngles(array([np.ravel(Arr_Theo2Theta),
                                        signchi * np.ravel(Arr_TheoChi)]).T,
                                        sorted_data)

        print('table_dist.shape', table_dist.shape)

    #         print "table_dist normal", table_dist[:5, :5]
    #     print "table_dist_new", table_dist.shape

    # shape is: along i   dim =  NBMAXSPOTS*nbmatrices = dim(theodata)
    #           along j   dim = dim(sorted_data)  experimental

    # prox_table = getArgmin(table_dist)

    # print "shape table_dist",shape(table_dist)
    # table_dist has shape (len(theo),len(exp))
    # tab[0] has len(exp) elements
    # tab[i] with i index of theo spot contains all distance from theo spot #i and all the exprimental spots
    # prox_table has len(exp) elements
    # prox_table[i] is the index of the exp. spot closest to theo spot (index i)
    # table_dist[i][prox_table[i]] is the minimum angular distance separating theo spot i from a exp spot of index prox_table[i]

    if verbose:
        #        print "/0", np.where(table_dist[:, 0] < 1.)
        #        print np.argmin(table_dist[:, 0])
        #        print "/4 exp", np.where(table_dist[:, 4] < 1.)
        #        print np.argmin(table_dist[:, 4])
        #        print np.shape(table_dist)
        print(table_dist[:, 4])

    #    pos_closest = np.transpose(array([arange(len(theodata)), prox_table]))
    #     nb_exp_spots = len(data_chi)
    #     nb_theo_spots = len(theodata)
    #     pos_closest_1d = array(arange(nb_theo_spots) * nb_exp_spots * ones(nb_theo_spots) + \
    #                                prox_table,
    #                                dtype=int32)
    #     allresidues = ravel(table_dist)[pos_closest_1d]

    allresidues = amin(table_dist, axis=1)
    taballresidues = allresidues.reshape((nbpeaksbunches, len(allresidues)//nbpeaksbunches))

    #     print "allresidues", allresidues
    # len(allresidues)  = len(theo)

    #     print "theodata", theodata
    #     print 'len(allresidues)', len(allresidues)
    if proxtable == 0:
        SIMILRATYTHRESHOLD = 0.9999
        nb_in_res = getNbMatches(taballresidues, SIMILRATYTHRESHOLD)

        # taballresiduesM=np.ma.masked_greater_equal(taballresidues,angtol)
        # meanres = np.mean(taballresiduesM,axis=1)
        # maxi = np.amax(taballresiduesM,axis=1)

        # res = allresidues[cond]
        # longueur_res = len(cond[0])
        # #         print allresidues
        # #         print "len(res)", len(res)
        # #         print "longueur_res", longueur_res
        # nb_in_res = len(res)
        # maxi = max(res)
        # meanres = mean(res)

        return taballresidues, taballresidues[0], nb_in_res, len(allresidues)
        # return taballresidues, taballresidues[0], nb_in_res, len(allresidues), meanres, maxi


def getNbMatches(residues, thresholdsimilarity):
    """return nb of good matches
    """
    cond = where(residues > thresholdsimilarity)
    return len(cond[0])
    # unique_elements, counts_elements = np.unique(cond[0], return_counts=True)
    # print("Frequency of unique values of the said array:")
    # print(np.asarray((unique_elements, counts_elements)))
    # nb_in_res=counts_elements
    # return np.sum(counts_elements)

def getProximity(TwicethetaChi,
                    data_theta,
                    data_chi,
                    angtol=0.5,
                    proxtable=0,
                    verbose=0,
                    signchi=1,
                    usecython=USE_CYTHON):
    r"""
    :param TwicethetaChi: (simulated or theoretical) two arrays of 2theta array and chi array (same length!)
    :param data_theta: array of theta angles (of experimental spots)
    :param data_chi: array of chi (same length!)

    :returns:  if proxtable = 1 : proxallresidues, res, nb_in_res, len(allresidues), meanres, maxi

    .. warning:: TwicethetaChi contains 2theta instead of data_theta theta contains theta !

    .. todo::
        * change the input with 2theta angles pnly to avoid confusion
        * remove the option signchi = 1 fixed old convention

    """
    # theo simul data
    theodata = array([TwicethetaChi[0] / 2.0, signchi * TwicethetaChi[1]]).T
    # exp data
    sorted_data = array([data_theta, data_chi]).T

    #     table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)
    #     print "table_dist_old", table_dist[:5, :5]
    #     print "table_dist_old", table_dist[-5:, -5:]
    #     print "table_dist_old", table_dist.shape

    if not usecython:
        table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)

        # for cartesian distance only, crude approximate for kf_direction=Z>0
        # import scipy
        # table_dist = scipy.spatial.distance.cdist(sorted_data, theodata).T

    else:
        # TODO to be improved by not preparing array?
        array_twthetachi_theo = array([TwicethetaChi[0], TwicethetaChi[1]]).T
        array_twthetachi_exp = array([data_theta * 2.0, data_chi]).T

        #         print "flags", array_twthetachi_theo.flags
        #         print "flags", array_twthetachi_exp.flags
        table_dist = angulardist.calculdist_from_2thetachi(array_twthetachi_theo.copy(order="c"),
                                                            array_twthetachi_exp.copy(order="c"))

    #         print "table_dist from cython", table_dist[:5, :5]

    #     print "table_dist_new", table_dist[:5, :5]
    #     print "table_dist_new", table_dist[-5:, -5:]
    #     print "table_dist_new", table_dist.shape

    if proxtable == 1:
        prox_table = getArgmin(table_dist)
    # print "shape table_dist",shape(table_dist)
    # table_dist has shape (len(theo),len(exp))
    # tab[0] has len(exp) elements
    # tab[i] with i index of theo spot contains all distance from theo spot #i and all the exprimental spots
    # prox_table has len(theo) elements
    # prox_table[i] is the index of the exp. spot closest to theo spot (index i)
    # table_dist[i][prox_table[i]] is the minimum angular distance
    # separating theo spot i from a exp spot of index prox_table[i]

    if verbose:
        #        print "/0", np.where(table_dist[:, 0] < 1.)
        #        print np.argmin(table_dist[:, 0])
        #        print "/4 exp", np.where(table_dist[:, 4] < 1.)
        #        print np.argmin(table_dist[:, 4])
        #        print np.shape(table_dist)
        print(table_dist[:, 4])

    #    pos_closest = np.transpose(array([arange(len(theodata)), prox_table]))
    #     nb_exp_spots = len(data_chi)
    #     nb_theo_spots = len(theodata)
    #     pos_closest_1d = array(arange(nb_theo_spots) * nb_exp_spots * ones(nb_theo_spots) + \
    #                                prox_table,
    #                                dtype=int32)
    #     allresidues = ravel(table_dist)[pos_closest_1d]

    allresidues = amin(table_dist, axis=1)

    #     print "allresidues", allresidues

    # len(allresidues)  = len(theo)

    #     print "theodata", theodata
    #     print 'len(allresidues)', len(allresidues)
    if proxtable == 0:
        cond = where(allresidues < angtol)
        res = allresidues[cond]
        longueur_res = len(cond[0])
        #         print allresidues
        #         print "len(res)", len(res)
        #         print "longueur_res", longueur_res
        if longueur_res <= 1:
            nb_in_res = longueur_res
            maxi = -min(allresidues)
            meanres = -1
        else:
            nb_in_res = len(res)
            maxi = max(res)
            meanres = mean(res)

        return allresidues, res, nb_in_res, len(allresidues), meanres, maxi

    elif proxtable == 1:
        return allresidues, prox_table, table_dist


def getProximity_new(Twicetheta, Chi, data_theta, data_chi,
                                                    angtol=0.5, proxtable=0,
                                                    verbose=0, signchi=1,
                                                    usecython=USE_CYTHON):
    """
    see doc of getProximity()

    """
    # theo simul data
    theodata = array([Twicetheta / 2.0, signchi * Chi]).T
    # exp data
    sorted_data = array([data_theta, data_chi]).T

    #     table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)
    #     print "table_dist_old", table_dist[:5, :5]
    #     print "table_dist_old", table_dist[-5:, -5:]
    #     print "table_dist_old", table_dist.shape

    if not usecython:
        table_dist = GT.calculdist_from_thetachi(sorted_data, theodata)
    #         print "table_dist", table_dist[:5, :5]
    else:
        # TODO to be improved by not preparing array
        import angulardist

        #         print "using cython"
        array_twthetachi_theo = array([Twicetheta, Chi]).T
        array_twthetachi_exp = array([data_theta * 2.0, data_chi]).T

        #         print "flags", array_twthetachi_theo.flags
        #         print "flags", array_twthetachi_exp.flags
        table_dist = angulardist.calculdist_from_2thetachi(array_twthetachi_theo.copy(order="c"),
                                                            array_twthetachi_exp.copy(order="c"))

    prox_table = getArgmin(table_dist)
    # print "shape table_dist",shape(table_dist)
    # table_dist has shape (len(theo),len(exp))
    # tab[0] has len(exp) elements
    # tab[i] with i index of theo spot contains all distance from theo spot #i and all the exprimental spots
    # prox_table has len(theo) elements
    # prox_table[i] is the index of the exp. spot closest to theo spot (index i)
    # table_dist[i][prox_table[i]] is the minimum angular distance separating theo spot i from a exp spot of index prox_table[i]

    if verbose:
        #        print "/0", np.where(table_dist[:, 0] < 1.)
        #        print np.argmin(table_dist[:, 0])
        #        print "/4 exp", np.where(table_dist[:, 4] < 1.)
        #        print np.argmin(table_dist[:, 4])
        #        print np.shape(table_dist)
        print(table_dist[:, 4])

    #    pos_closest = np.transpose(array([arange(len(theodata)), prox_table]))
    #     nb_exp_spots = len(data_chi)
    #     nb_theo_spots = len(theodata)
    #     pos_closest_1d = array(arange(nb_theo_spots) * nb_exp_spots * ones(nb_theo_spots) + \
    #                                prox_table,
    #                                dtype=int32)
    #     allresidues = ravel(table_dist)[pos_closest_1d]

    allresidues = amin(table_dist, axis=1)

    # len(allresidues)  = len(theo)

    #     print "theodata", theodata
    #     print 'len(allresidues)', len(allresidues)
    if proxtable == 0:
        cond = where(allresidues < angtol)
        res = allresidues[cond]
        longueur_res = len(cond[0])
        # print allresidues
        #         print "len(res)", len(res)
        #         print "longueur_res", longueur_res
        if longueur_res <= 1:
            nb_in_res = longueur_res
            maxi = -min(allresidues)
            meanres = -1
        else:
            nb_in_res = len(res)
            maxi = max(res)
            meanres = mean(res)

        return allresidues, res, nb_in_res, len(allresidues), meanres, maxi

    elif proxtable == 1:
        return allresidues, prox_table, table_dist


def Angular_residues_np(test_Matrix, twicetheta_data, chi_data, ang_tol=0.5,
                                                                key_material="Si",
                                                                emin=5,
                                                                emax=25,
                                                                ResolutionAngstrom=False,
                                                                detectorparameters=None,
                                                                onlyXYZ=False,
                                                                simthreshold=0.999,
                                                                dictmaterials=dict_Materials):
    r"""
    Computes angular residues between pairs of close exp. and
    theo. spots simulated according to test_Matrix, within tolerance angle

    .. note::
        * used in manual indexation
        * Used in AutoIndexation module
        * Used in FileSeries

    :param twicetheta_data: experimental 2theta angles of scattered beams or spots (kf vectors)
    :type twicetheta_data: array
    :param chi_data: experimental chi angles of scattered beams or spots (kf vectors)
    :param test_Matrix: Orientation matrix
    :type test_Matrix: 3x3 array
    :param ang_tol: angular tolerance in degrees to accept or reject a exp. and theo pair
    :type ang_tol: scalar

    :param detectorparameters: Dictionary of detector parameters (key, value) that must constain:
                            'kf_direction' , general position of detector plane
                            'detectordistance', detector distance (mm)
                            'detectordiameter', detector diameter (mm)
                            'pixelsize' and 'dim'
    """
    if detectorparameters is None:
        # use default parameter
        kf_direction = "Z>0"
        detectordistance = 70.0
        detectordiameter = 165.0
        pixelsize = 165.0 / 2048
        dim = (2048, 2048)
    else:
        kf_direction = detectorparameters["kf_direction"]
        detectordistance = detectorparameters["detectorparameters"][0]
        detectordiameter = detectorparameters["detectordiameter"]
        pixelsize = detectorparameters["pixelsize"]
        dim = detectorparameters["dim"]

    #     print "kf_direction,pixelsize,detectordistance", kf_direction, pixelsize, detectordistance

    # ---simulation-----------------------------------
    grain = CP.Prepare_Grain(key_material, test_Matrix, dictmaterials=dictmaterials)

    # array(vec) and array(indices) (here with fastcompute=0 array(indices)=0)
    # of spots exiting the crystal towards detector plane and geometry given by kf_direction
    spots2pi = LAUE.getLaueSpots(CST_ENERGYKEV / emax, CST_ENERGYKEV / emin,
                                    [grain],
                                    fastcompute=1,
                                    verbose=0,
                                    kf_direction=kf_direction,
                                    ResolutionAngstrom=ResolutionAngstrom,
                                    dictmaterials=dictmaterials)

    if not SCIKITLEARN or not onlyXYZ:
        # 2theta,chi of spot which are on camera (with harmonics)
        # None because no need of hkl vectors
        # TwicethetaChi without energy calculations and hkl selection
        # without use of spots instantation (faster)
        TwicethetaChi = LAUE.filterLaueSpots_full_np(spots2pi[0][0], None, onlyXYZ=False,
                                                        HarmonicsRemoval=0,
                                                        fastcompute=1,
                                                        kf_direction=kf_direction,
                                                        detectordistance=detectordistance,
                                                        detectordiameter=detectordiameter,
                                                        pixelsize=pixelsize,
                                                        dim=dim)

        # old calculation with spots instantiation
        #     TwicethetaChi = LAUE.filterLaueSpots(spots2pi, fileOK=0, fastcompute=1,
        #                                          kf_direction=kf_direction,
        #                                          detectordistance=detectordistance,
        #                                          detectordiameter=detectordiameter,
        #                                          pixelsize=pixelsize,
        #                                          dim=dim)

        #     print "len(TwicethetaChi[0])", len(TwicethetaChi[0])
        if len(TwicethetaChi[0]) == 0:
            #         print 'no peak found'
            return None

        # no particular gain...?
        return getProximity(TwicethetaChi, twicetheta_data / 2.0, chi_data, angtol=ang_tol,
                                                                                        proxtable=0)

    else:
        Q_XYZ_onCam = LAUE.filterLaueSpots_full_np(spots2pi[0][0], None, onlyXYZ=True,
                                                    HarmonicsRemoval=0,
                                                    fastcompute=1,
                                                    kf_direction=kf_direction,
                                                    detectordistance=detectordistance,
                                                    detectordiameter=detectordiameter,
                                                    pixelsize=pixelsize,
                                                    dim=dim)

        # Y should be Q vectors corresponding to exp. twicetheta_data and chi_data
        Y = LaueGeo.from_twchi_to_q((twicetheta_data, chi_data)).T

        # print("Q_XYZ_onCam   theo",Q_XYZ_onCam)
        # print("Y exp.",Y)

        # print("Q_XYZ_onCam   theo  shape",Q_XYZ_onCam.shape)
        # print("Y exp.  shape",Y.shape)
        # return ssd.cosine(Q_XYZ_onCam, Y)
        # return np.arccos(1-sm.pairwise.cosine_similarity(Q_XYZ_onCam, Y, dense_output=True))*180./np.pi
        smMat = sm.pairwise.cosine_similarity(Q_XYZ_onCam, Y, dense_output=True)
        SIMILRATYTHRESHOLD = simthreshold
        nb_in_res = getNbMatches(smMat, SIMILRATYTHRESHOLD)
        return nb_in_res


def Angular_residues_np_multimatrices(ListMatrices, twicetheta_data, chi_data, ang_tol=0.5,
                                                                    key_material="Si",
                                                                    emin=5,
                                                                    emax=25,
                                                                    ResolutionAngstrom=False,
                                                                    detectorparameters=None,
                                                                    dictmaterials=dict_Materials):

    """ See doc of Angular_residues_np()

    """
    NBMAXPEAKS = 1700

    if detectorparameters is None:
        # use default parameter
        kf_direction = "Z>0"
        detectordistance = 70.0
        detectordiameter = 165.0
        pixelsize = 165.0 / 2048
        dim = (2048, 2048)
    else:
        kf_direction = detectorparameters["kf_direction"]
        detectordistance = detectorparameters["detectorparameters"][0]
        detectordiameter = detectorparameters["detectordiameter"]
        pixelsize = detectorparameters["pixelsize"]
        dim = detectorparameters["dim"]

    #     print "kf_direction,pixelsize,detectordistance", kf_direction, pixelsize, detectordistance

    # ---simulation

    nbmatrices = len(ListMatrices)

    Arr_Theo2Theta = np.empty((nbmatrices, NBMAXPEAKS))
    Arr_TheoChi = np.empty((nbmatrices, NBMAXPEAKS))
    # print "Reference Element or structure label", key_material
    for matindex in list(range(nbmatrices)):
        test_Matrix = ListMatrices[matindex]
        grain = CP.Prepare_Grain(key_material, test_Matrix)
        # array(vec) and array(indices) (here with fastcompute=0
        # array(indices)=0) of spots exiting the crystal in 2pi steradian (Z>0)
        spots2pi = LAUE.getLaueSpots(CST_ENERGYKEV / emax,
                                    CST_ENERGYKEV / emin,
                                    [grain],
                                    fastcompute=1,
                                    verbose=0,
                                    kf_direction=kf_direction,
                                    ResolutionAngstrom=ResolutionAngstrom,
                                    dictmaterials=dictmaterials)

        # 2theta,chi of spot which are on camera (with harmonics)
        # None because no need of hkl vectors
        # TwicethetaChi without energy calculations and hkl selection
        # without use of spots instantation (faster)
        TheoTwicethetaChi = LAUE.filterLaueSpots_full_np(spots2pi[0][0], None, HarmonicsRemoval=0,
                                                        fastcompute=1,
                                                        kf_direction=kf_direction,
                                                        detectordistance=detectordistance,
                                                        detectordiameter=detectordiameter,
                                                        pixelsize=pixelsize,
                                                        dim=dim)
        # print("matindex",matindex)
        Arr_Theo2Theta[matindex] = TheoTwicethetaChi[0][:NBMAXPEAKS]
        Arr_TheoChi[matindex] = TheoTwicethetaChi[1][:NBMAXPEAKS]
        # if len(TheoTwicethetaChi[0]) == 0:
        #     #         print 'no peak found'
        #     return None

    # no particular gain...?
    return getProximity_multimatrices(Arr_Theo2Theta,
                                    Arr_TheoChi, twicetheta_data / 2.0,
                                    chi_data,
                                    angtol=ang_tol,
                                    proxtable=0)


def Angular_residues(test_Matrix, twicetheta_data, chi_data, ang_tol=0.5, key_material="Si",
                                emin=5,
                                emax=25,
                                ResolutionAngstrom=False,
                                detectorparameters=None,
                                dictmaterials=dict_Materials):

    """ see doc of Angular_residues_np()

    """
    if detectorparameters is None:
        # use default parameter
        kf_direction = "Z>0"
        detectordistance = 70.0
        detectordiameter = 165.0
        pixelsize = 165.0 / 2048
        dim = (2048, 2048)
    else:
        kf_direction = detectorparameters["kf_direction"]
        detectordistance = detectorparameters["detectorparameters"][0]
        detectordiameter = detectorparameters["detectordiameter"]
        pixelsize = detectorparameters["pixelsize"]
        dim = detectorparameters["dim"]

    #     print "kf_direction,pixelsize,detectordistance", kf_direction, pixelsize, detectordistance

    # ---simulation
    # print "Reference Element or structure label", key_material

    # spots2pi = generalfabriquespot_fromMat_veryQuick(CST_ENERGYKEV/emax,CST_ENERGYKEV/emin,[grain],1,fastcompute=1,fileOK=0,verbose=0)
    grain = CP.Prepare_Grain(key_material, test_Matrix)

    # array(vec) and array(indices) (here with fastcompute=0 array(indices)=0) of spots exiting the crystal in 2pi steradian (Z>0)
    spots2pi = LAUE.getLaueSpots(CST_ENERGYKEV / emax,
                                    CST_ENERGYKEV / emin,
                                    [grain],
                                    fastcompute=1,
                                    verbose=0,
                                    kf_direction=kf_direction,
                                    ResolutionAngstrom=ResolutionAngstrom,
                                    dictmaterials=dictmaterials)
    # 2theta,chi of spot which are on camera (with harmonics)
    TwicethetaChi = LAUE.filterLaueSpots(spots2pi,
                                        fileOK=0,
                                        fastcompute=1,
                                        kf_direction=kf_direction,
                                        detectordistance=detectordistance,
                                        detectordiameter=detectordiameter,
                                        pixelsize=pixelsize,
                                        dim=dim)

    #     print "len(TwicethetaChi[0])", len(TwicethetaChi[0])
    #     print "len(TwicethetaChi[0])", len(TwicethetaChi[0])
    if len(TwicethetaChi[0]) == 0:
        return None

    return getProximity(TwicethetaChi, twicetheta_data / 2.0, chi_data, angtol=ang_tol, proxtable=0)


def getMatchingRate(indexed_spots_dict, test_Matrix, ang_tol, simulparam, removeharmonics=1,
                                                                            detectordiameter=165.0):
    r"""
    Gets matching rate for an orientation matrix
                    for all exp. data stored in dictionary indexed_spots_dict

    input:
    :param indexed_spots_dict: dict of exp spots (indexed or not)
    :param test_Matrix: orientation matrix (3x3)
    :param ang_tol: angular tolerance below which a pair between
                            exp. and simulated spots is accepted
    :param simulparam: tuple containing emin, emax, key_material, detectorparameters
    :param removeharmonics: 0 or 1 to compute respectively without or with harmonics in simulated pattern

    :returns: * [0], number of pairs between exp and theo spots
                * [1], number of theo. spots
                * [2], matching rate (ratio*100 of the two previous numbers)
    """
    emin, emax, key_material, detectorparameters = simulparam
    # TODO to shorten
    _, twicetheta_data, chi_data, _, _, intensity_data = ISS.getSpotsData(indexed_spots_dict).T

    # simulated data
    grain = CP.Prepare_Grain(key_material, test_Matrix)
    (Twicetheta, Chi, Miller_ind,
    _, _, Energy) = LAUE.SimulateLaue(grain,
                                            emin,
                                            emax,
                                            detectorparameters,
                                            removeharmonics=removeharmonics,
                                            detectordiameter=detectordiameter * 1.25)

    nb_of_simulated_spots = len(Twicetheta)

    # find close pairs between exp. and theo. spots
    res = SpotLinks(twicetheta_data,
                    chi_data,
                    intensity_data,  # experimental data
                    ang_tol,  # tolerance angle
                    Twicetheta,
                    Chi,
                    Miller_ind,
                    Energy,
                    absoluteindex=None)

    if res == 0 or len(res[1]) == 0:
        return None, nb_of_simulated_spots, None

    print("res_links", res[1])

    print("nb of exp spots data probed :", len(twicetheta_data))

    nb_of_links = len(res[1])

    matching_rate = 100.0 * nb_of_links / nb_of_simulated_spots

    return nb_of_links, nb_of_simulated_spots, matching_rate


def getStatsOnMatching(List_of_Angles,
                    twicetheta_data,
                    chi_data,
                    key_material,
                    ang_tol=1.0,
                    emin=5,
                    emax=25,
                    intensity_data=None,
                    verbose=0):
    """
    return matching rate in terms of nb of close pairs of spots (exp. and simul. ones)
    within angular tolerance
    (simulated pattern contains harmonics)

    :param List_of_Angles: list of 3 angles (EULER angles)
    :type List_of_Angles: list of 3 floats
    :param twicetheta_data, chi_data: experimental spots angles coordinates (kf vectors)
    :type twicetheta_data, chi_data: 2 arrays of floats
    :param key_material: key for material used in simulation
    :type key_material: string
    :param ang_tol: matching angular tolerance to form pairs
    :type ang_tol: float

    .. note:: USED in filterEulersList()
    """
    kk = 0
    allmatchingrate = []
    for angles_sol in List_of_Angles:

        test_Matrix = GT.fromEULERangles_toMatrix(angles_sol)

        res = Angular_residues(test_Matrix,
                                twicetheta_data,
                                chi_data,
                                ang_tol=ang_tol,
                                key_material=key_material,
                                emin=emin,
                                emax=emax)

        if res is None:
            continue

        matching_rate = 100.0 * res[2] / res[3]
        # absolute_matching_rate = res[2]
        allmatchingrate.append(matching_rate)

        if verbose:
            print("*" * 30)
            print(
                "res for k:%d and angles: [%.1f,%.1f,%.1f]"
                % (kk, angles_sol[0], angles_sol[1], angles_sol[2]))
            #        print "res",res[2:]
            print("Matching rate ----(in %%):                %.2f " % matching_rate)
            print("Nb of close spot pairs (< %.2f deg) : %d" % (ang_tol, res[2]))
            print("Nb of simulated spots : %d" % res[3])
            print("mean angular residues %.2f deg." % res[4])
            print("highest residues %.2f deg." % res[5])
        kk += 1

    return np.argsort(np.array(allmatchingrate))[::-1], allmatchingrate


if __name__ == "__main__":
    import time

    npoints1 = 60  # theo
    npoints2 = 120  # exp

    listpoints1 = np.random.random_sample((npoints1, 2)) * (180.0) - 90.0
    listpoints2 = np.random.random_sample((npoints2, 2)) * (180.0) - 90.0

    # listpoints1.dtype = numpy.double
    # listpoints2.dtype = numpy.double

    listpoints1_theta = listpoints1
    listpoints2_theta = listpoints2

    listpoints1_theta[:, 0] = listpoints1[:, 0] / 2.0
    listpoints2_theta[:, 0] = listpoints2[:, 0] / 2.0
    # theo
    TwicethetaChi1 = listpoints1_theta[:, 0], listpoints1_theta[:, 1]

    Twicetheta1, Chi1 = listpoints1_theta[:, 0], listpoints1_theta[:, 1]
    # exp.
    Twicetheta2 = listpoints2_theta[:, 0]
    Chi2 = listpoints2_theta[:, 1]

    inittime = time.time()
    res = getProximity(TwicethetaChi1, Twicetheta2, Chi2, angtol=0.5, proxtable=0)
    finaltime = time.time()

    time_old = finaltime - inittime

    print("computation time old is ", time_old)

    inittime = time.time()
    res = getProximity(TwicethetaChi1, Twicetheta2, Chi2, angtol=0.5, proxtable=0, usecython=False)
    finaltime = time.time()

    time_new = finaltime - inittime

    print("computation time new is ", time_new)

    inittime = time.time()
    res = getProximity(TwicethetaChi1, Twicetheta2, Chi2, angtol=0.5, proxtable=0, usecython=True)

    #     res = getProximity_new(Twicetheta1, Chi1,
    #                               Twicetheta2, Chi2,
    #                               angtol=.5, proxtable=0,
    #                               usecython=True)
    finaltime = time.time()

    time_newc = finaltime - inittime

    print("computation time new + cython is ", time_newc)

    print("ratio old/new", time_old / time_new)
    print("ratio old/new +cython", time_old / time_newc)



"""
Module to deduce from distances recognition in mutual pair distance matrix
the adjaceny matrix from which graph theory's algorithms can produce
set of spots highly connected to each other, hence spos likely
to belong to the same grain

Tabledistance comes from most important planes (and intense reflection) in cubic structure up to family plane (321)

TODO: take into account the information related to the connectivity between two spots
"""

import sys
import os
# import scipy.io.array_import # pour charger les donnees
import pickle

import numpy as np

import pylab as P

sys.path.insert(0, os.path.join(os.path.dirname(__file__), os.path.pardir))

try:
    import networkx as NX
except ImportError:
    print("\n***********************************************************")
    print("networkx module is missing! Please install it !")
    print("***********************************************************\n")

if sys.version_info.major == 3:
    from . import findorient as FO
    # from . import indexingAnglesLUT as INDEX
    from . import generaltools as GT
    from . import IOLaueTools as IOLT
else:
    import findorient as FO
    # import indexingAnglesLUT as INDEX
    import generaltools as GT
    import IOLaueTools as IOLT

__author__ = "Jean-Sebastien Micha, CRG-IF BM32 @ ESRF"


def read_LUT(picklefilename):

    return pickle.load(open(picklefilename, 'rb'))


def create_tabdist(pickle_it=0, picklefilename="CubicDistanceList.dat"):
    """
    Creates all distances below 90 degrees between two planes in CUBIC non distorted structure
    from main planes (1,0,0),(1,1,0) up to (3,2,1)  type

    options pickle_it=1 to save table in picklefilename
    """

    whole_interdistance_list = []
    for func in FO.LUT_MAIN_CUBIC:
        whole_interdistance_list = whole_interdistance_list + list(func)

    # print np.sort(array(whole_interdistance_list))

    ar_distances = np.sort(np.array(np.array(list(
                    set(list(np.array(np.sort(np.array(whole_interdistance_list)).round(
                                    decimals=3), dtype="|S6"))))), dtype=np.float32))
    print("Table of distances")
    print(ar_distances)

    if pickle_it:
        distfile = open(picklefilename, "w")
        pickle.dump(ar_distances, distfile, protocol=2)
        # binary mode , data were created with cPickle.dump(obj,filenprotocol=2)
        # frou=open(Globalname,'rb')
        # tabdata=cPickle.load(frou)
        distfile.close()
        print("Table pickled in %s" % picklefilename)
    return ar_distances


def read_disttable(picklefilename=None):
    """
    Reads all sorted distances in cubic structure below 90 degrees
    from main planes (1,0,0),(1,1,0) up to (3,2,1)  type
    """
    if picklefilename is not None:
        ar_distances = read_LUT(picklefilename)
        print("Cubic distance table read from a pickled file %s"%picklefilename)
    else:
        ar_distances = np.array([8.13000011, 10.02499962, 10.89299965, 11.48999977, 14.76299953, 15.79300022, 17.02400017,
        17.54800034, 17.71500015, 18.43499947, 19.10700035, 19.2859993, 19.47100067, 21.61700058, 21.78700066,
        22.20800018, 24.09499931, 25.23900032, 25.35199928, 25.84199905, 26.56500053, 27.0170002,
        27.26600075, 29.20599937, 29.49600029, 30.0, 31.00300026, 31.48200035, 31.94799995, 32.31200027,
        32.51300049, 33.21099854, 33.55699921, 35.09700012, 35.26399994, 36.31000137, 36.69900131, 36.86999893,
        38.94200134, 39.23199844, 40.20299911, 40.29100037, 40.47900009, 40.89300156, 41.81000137,
        42.39199829, 42.45000076, 43.0890007, 44.41500092, 45.0, 45.28900146, 47.12400055, 47.45899963,
        47.60800171, 47.86999893, 48.18999863, 49.10699844, 49.79700089, 49.86000061, 49.99499893, 50.47900009,
        50.76800156, 51.88700104, 53.13000107, 53.30099869, 53.39599991, 53.72900009, 54.73600006, 55.10499954,
        55.4620018, 56.78900146, 56.93799973, 57.68799973, 58.19400024, 58.51800156, 58.9090004, 59.52999878,
        59.83300018, 60.0, 60.50400162, 61.08599854, 61.43899918, 61.87400055, 62.9640007, 63.43500137,
        63.54899979, 63.61199951, 64.6230011, 64.76100159, 64.89600372, 65.00299835, 65.06199646, 65.90499878,
        66.13899994, 66.42199707, 67.58000183, 67.79199982, 68.58300018, 68.98799896, 69.07499695, 70.52899933,
        70.89299774, 71.19599915, 71.56500244, 72.02500153, 72.45200348, 72.54199982, 72.65399933, 73.22100067,
        73.39800262, 73.56999969, 74.20700073, 74.49900055, 75.03700256, 75.31300354, 75.7480011, 76.36699677,
        77.07900238, 77.39600372, 78.46299744, 78.90399933, 79.00700378, 79.10700226, 79.48000336, 79.73699951,
        79.97499847, 80.40599823, 80.72599792, 81.78700256, 81.87000275, 82.17900085, 82.25099945, 82.58200073,
        83.13500214, 83.6210022, 83.9489975,
        84.23200226, 84.26100159, 84.78399658, 84.88899994, 85.15200043, 85.90399933, 90.0], dtype=float)
        print("Cubic distances Table read")

        # shrink LUT
        # ar_distances =  ar_distances[ar_distances>=55.]
    return ar_distances


def TableDistance_exp(filename, nb_of_spots, col_Int=4):
    """
    From experiments file and number of first selected spots
    Returns matrix of mutual angular distances (deg)

    nb_of_spots=-1 means considering all spots
    Uses intensity sorting from data in column index col_Int to select the first spots

    """
    #(alldata, data_theta, data_chi, data_pixX, data_pixY, data_I, detParam)
    data = IOLT.readfile_cor(filename)
    data_theta, data_chi, data_I = data[1], data[2], data[5]
    nbp = len(data_theta)

    if nb_of_spots > 0:
        upto = min(nb_of_spots, nbp)
        sorted_int_index = np.argsort(data_I)[::-1][:upto]
        print("Considering only %d most intense spots" % upto)
    elif nb_of_spots == -1:
        print("Considering all spots")
        sorted_int_index = np.argsort(data_I)[::-1]
        upto = nbp

    # listofselectedpts = np.arange(len(sorted_int_index))
    Theta = data_theta[sorted_int_index]
    Chi = data_chi[sorted_int_index]
    Intens = data_I[sorted_int_index]
    sorted_data = np.transpose(np.array([Theta, Chi, Intens]))

    # array of interangular distance of all points
    print("Calculating all angular distances ...")
    angulardisttable = GT.calculdist_from_thetachi(sorted_data[:, 0:2], sorted_data[:, 0:2])
    print("... Done !")
    # ind_sort=argsort(Tabledistance[0,1:])
    return angulardisttable, upto


def create_AdjencyMatrix(Tabledistance, ReferenceTable, ang_tol, nb_of_spots, verbose=0):
    """
    Creates Adjency Matrix from:
        Tabledistance: matrix of mutual inter angular distance
        ReferenceTable: table of possible distances
        ang_tol: tolerance for assigning 1 in adjencymatrix if exp. distance "is" in ReferenceTable

    nb_of_spots=len(Tabledistance)
    """
    mytab = np.zeros(nb_of_spots * nb_of_spots)
    print("creating adjency matrix ...")

    # put 1 where a distance was found in ReferenceTable for each row = elem

    # for elem in arange(nb_of_spots):
    # print elem
    # print GT.find_closest(ReferenceTable ,Tabledistance[elem],ang_tol)[1]
    # print GT.find_closest(ReferenceTable ,Tabledistance[elem],ang_tol)[1]+elem*nb_of_spots

    # enthought python 2.4
    # john=map(lambda elem: mytab.put(1,    GT.find_closest(ReferenceTable ,Tabledistance[elem],ang_tol)[1]  +elem*nb_of_spots),  arange(nb_of_spots) )
    # enthought python 2.5 (reversed order of args!!)
    adjtab1D = [mytab.put(
            GT.find_closest(ReferenceTable, Tabledistance[elem], ang_tol)[1] + elem * nb_of_spots, 1)
            for elem in np.arange(nb_of_spots)]

    # for debugging
    # mytab=np.zeros(nb_of_spots*nb_of_spots)
    # print "ReferenceTable",ReferenceTable
    # for elem in np.arange(nb_of_spots):
    # print "elem",elem
    # print "angkl",Tabledistance[elem]
    # posy=GT.find_closest(Tabledistance[elem],ReferenceTable ,ang_tol)[1]
    # print "posy raw"
    # posy=posy+elem*nb_of_spots
    # what=1
    # print  "posy",posy
    # mytab.put(posy,what)

    adjencymat = np.reshape(mytab, (nb_of_spots, nb_of_spots))
    if verbose:
        print("adjtab1D", adjtab1D)
    print("... Done !")

    """
    # to understand the creation of john!
    resclose=GT.find_closest(ar_distances,Tabledistance[0],0.1)
    #print resclose[0]
    print resclose[1]
    #print resclose[2]
    #print "ar_distances input",ar_distances
    #print "Tabledistance[0] target",Tabledistance[0]
    # resclose[1] indice dans sort(Tabledistance[0,1:]) qui ont une distance connue

    #locations ----------
    indices_in_ref=resclose[0][resclose[1]]
    indices_in_data=resclose[1]
    print "these two list must be equal within the angular tolerance"
    print take(ar_distances,resclose[0][resclose[1]])
    print take(Tabledistance[0],resclose[1])
    # ------------------

    # ------------------
    #print "bigmapping"
    #bozo=map(lambda elem: GT.find_closest(ar_distances,sort(Tabledistance[elem]),0.1)[1],range(200))
    # ------------------
    """
    return adjencymat

def flatnestedlist(list_of_lists):
    return [y for x in list_of_lists for y in x]

def bestclique_oneNode(CliquesList, displaybest=0):
    """
    Returns the largest (in size) clique containing the spot cliques list the largest one

    """
    # list of clique length
    Clength_list = np.array(list(map(len, CliquesList)))
    # print('Clength_list',Clength_list)

    # index to sort Cliques by length
    sortedcliques_ind = np.argsort(Clength_list)[::-1]

    # print searchsorted(lon_C0[sortedcliques_ind],4)
    print(CliquesList[np.argmax(Clength_list)])
    bestclique = np.sort(CliquesList[np.argmax(Clength_list)])

    val_longest = np.amax(Clength_list)
    longest_inds = np.where(Clength_list == val_longest)[0]
    longest_Cliques = [CliquesList[ind] for ind in longest_inds]
    spots_in_longest = set(flatnestedlist(longest_Cliques))
    print("longest_Cliques", longest_Cliques)
    print("spots_in_longest", spots_in_longest)

    # val_threshold = 8
    # higher_inds = np.where(Clength_list>=val_threshold)[0]
    # higher_Cliques = [CliquesList[ind] for ind in higher_inds]
    # spots_in_higher = set(flatnestedlist(higher_Cliques))
    # print("higher_Cliques",higher_Cliques)
    # print("spots_in_higher",spots_in_higher)

    statlongest = np.bincount(np.ravel(longest_Cliques))
    highfreqspot = np.argsort(statlongest)[::-1]
    print("Most frequent spots ", statlongest[highfreqspot][:10])
    print(" with frequency ", highfreqspot[:10])

    print("bestclique", bestclique)
    if displaybest:
        nbbest_r = displaybest  # to select the number of best cliques found
        for index in sortedcliques_ind[:nbbest_r]:
            print("clique size", Clength_list[index], "   nodes: ", np.sort(CliquesList[index]))
    return bestclique


def give_bestclique(filename, nb_of_spots, ang_tol, nodes=0, col_Int=-1,
                                        LUTfilename=None, verbose=0):
    """ from peakslist file, it gives the sets of spots belonging to cliques according to a structure (given bu the LUT)

    if LUTfilename is None (default) then cubic structure is given

    .. note:: spot indices are those of file (whatever some spots have been already indexed)
    """
    # reading reference table of distances for recognition
    ar_distances = read_disttable(picklefilename=LUTfilename)
    # print "ar_distances",ardistances

    # reading experimental data i.e. list of spots from .cor file (2the,chi,x,y,I)
    if verbose:
        print("data file: %s" % filename)
    Tabledistance, nb_of_spots = TableDistance_exp(filename, nb_of_spots, col_Int=col_Int)

    # print "Tabledistance in give_bestclique [:10]",Tabledistance[:10]
    # print "nb_of_spots",nb_of_spots

    # Adjency matrix creation and corresponding graph creation:
    adjencymat = create_AdjencyMatrix(Tabledistance, ar_distances, ang_tol, nb_of_spots)
    # GGraw = NX.from_whatever(adjencymat, create_using=NX.Graph()) # old syntax
    GGraw = NX.to_networkx_graph(adjencymat, create_using=NX.Graph())

    # print adjencymat
    # print shape(adjencymat)
    # print adjencymat[0]

    print("nodes selected", nodes)
    print("Searching cliques ...")
    C0 = NX.cliques_containing_node(GGraw, nodes=nodes, cliques=None)
    print('\nC0', C0)
    print("\n... Done !\n")
    if not isinstance(nodes, int):
        best_list = []
        print("nb of nodes", len(C0))
        for k in range(len(C0)):
            if verbose:
                print("finding best cliques for # ", k, " entered nodes")
            bc = bestclique_oneNode(C0[k], displaybest=verbose)
            if verbose:
                print("----------------")
            best_list.append(np.sort(bc))
    else:
        # single node results
        best_list = np.sort(bestclique_oneNode(C0, displaybest=verbose))
    return best_list


def test(plot=0):
    """
    Test with a simulated noisy adjency matrix containing two grains
    """
    index_noise = 5
    index_connectivity = 0

    # noise=poisson(lam=2,size=(20,20))/4
    noise = np.clip(np.random.random_integers(-index_noise, 1, (20, 20)), 0, 1)

    nbflip = 100.0 * np.bincount(np.ravel(noise))[1] / np.bincount(np.ravel(noise))[0]
    print("rate of noise error", nbflip)
    a1 = np.reshape(np.clip(np.random.random_integers(-index_connectivity, 1, size=100), 0, 1),
        (10, 10))
    a2 = np.reshape(np.clip(np.random.random_integers(-index_connectivity, 1, size=100), 0, 1),
        (10, 10))
    a0 = np.zeros((10, 10))
    bigmat = np.array(np.vstack((np.hstack((a1, a0)), np.hstack((a0, a2)))), dtype=np.int8)
    print("bigmat", bigmat)
    print("noise", noise)

    # GGraw = NX.from_whatever(bigmat, create_using=NX.Graph()) #old syntax
    # GGraw = NX.to_networkx_graph(bigmat, create_using=NX.Graph())
    # GGnoise = NX.from_whatever(np.bitwise_xor(bigmat, noise), create_using=NX.Graph()) #old syntax
    GGnoise = NX.to_networkx_graph(np.bitwise_xor(bigmat, noise), create_using=NX.Graph())

    # cliques of noisy data
    Listcliques_noise = [cli for cli in NX.find_cliques(GGnoise)]
    lon_noise = np.array(list(map(len, Listcliques_noise)))

    print("noise", lon_noise)
    print("noise", np.argsort(lon_noise)[::-1])

    nbbest = 10
    for index in np.argsort(lon_noise)[::-1][:nbbest]:
        print("clique size", lon_noise[index], "   nodes: ", Listcliques_noise[index])

    # # cliques of perfect data
    # Listcliques_r=NX.find_cliques(GGraw)
    # lon_r=np.array(map(len,Listcliques_r))

    # print "raw",lon_r
    # print "raw",np.argsort(lon_r)[::-1]

    # nbbest_r=10
    # for index in np.argsort(lon_r)[::-1][:nbbest_r]:
    # print "clique size",lon_r[index],"   nodes: ",Listcliques_r[index]

    if plot:
        NX.draw(GGnoise)
        P.show()
    print("end of test")


def test_findcliques():
    """
    Test for finding highly connected set of spots
    Connection exists between two spots if their mutual angular distance is found in lookup table
    """

    nb_of_spots = -1  # nb of first most intense peaks
    ang_tol = 0.05
    # filename='fiveUO2grains.cor'
    filename = "Zr_A169_0220.cor"
    # filename='Ge_Zr_10000.cor'

    print("************\n For spots 0\n ***************")
    NODES = 0
    give_bestclique(filename, nb_of_spots, ang_tol, nodes=NODES, col_Int=4, verbose=0)

    print("\n******************\nFor the ten next most intense spots\n********************")
    NODES = list(range(1, 11))
    give_bestclique(filename, nb_of_spots, ang_tol, nodes=NODES, col_Int=4, verbose=0)


# ---------------------------------------------------
# -------- MAIN -------------------------------------
# ---------------------------------------------------


if __name__ == "__main__":

    # test_findcliques()
    #test(plot=1)
    """
    #print searchsorted(lon_C0[sortedcliques],4)
    bestclique=sort(C0[argmax(lon_C0)])
    print "bestclique",bestclique
    if 1:
        nbbest_r=200 #to select the number of best cliques found
        for index in sortedcliques[:nbbest_r]:
            print "clique size",lon_C0[index],"   nodes: ",sort(C0[index])
    """

    """
    # all cliques finding in graph
    print"finding cliques ----------------"
    Listcliques_r=find_cliques(GGraw)
    print "OK, that's done"
    lon_r=array(map(len,Listcliques_r))
    maxi=max(lon_r)
    print "raw",lon_r
    print "raw",argsort(lon_r)[::-1]

    nbbest_r=50 #to select the number of best cliques found
    for index in argsort(lon_r)[::-1][:nbbest_r]:
        print "clique size",lon_r[index],"   nodes: ",sort(Listcliques_r[index])


    """

    """
    NX.draw(GGraw)
    P.show()
    """

    """
    # in order to find sets corresponding to each grain -------------------
    listsortedcliques=[Listcliques_r[np.argsort(lon_r)[::-1][elem]] for elem in range(len(Listcliques_r))]
    Grains=NX.Graph()

    def update_graph(Graph,list_spot):

        if len(list_spot)>0:
            NX.Graph.add_nodes_from(list_spot)
            shift_r=list_spot[1:]+list_spot[:1]

            edges=np.transpose(np.array([list_spot,shift_r]))

            NX.Graph.add_edges_from(edges)

    def update_graph2(Graph,list_spot):
        lon=len(list_spot)

        if lon>0:
            #print "listspot",list_spot
            NX.Graph.add_nodes_from(list_spot)
            edges=np.take(array(list_spot),NX.complete_graph(lon).edges())

            NX.Graph.add_edges_from(edges)

    for m in range(6):
        print "**** ",m
        NX.update_graph2(Grains,listsortedcliques[m])



    NX.draw(Grains)
    P.show()

    """
